#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "head.h"

double fuv( eps, a, x, u, v )
	double eps, a, x, u, v ;
{
	return ( (u*(1.-u)*(u-a)-v)/eps + x ) ;
}

double guv( b, u, v )
	double b, u, v ;
{
	return ( u-b*v ) ;
}

int open_files( pps )
	PARAMS *pps ;
{
	char str[STR_LEN] ;

	if ( pps->sw_u == 1 ) {
		strcat( strcpy( str, pps->result_file ), ".u" ) ;
		if ( ( pps->fpu = fopen( str, "wb" ) ) == NULL )
			return -1 ;
	}

	if ( pps->sw_v == 1 ) {
		strcat( strcpy( str, pps->result_file ), ".v" ) ;
		if ( ( pps->fpv = fopen( str, "wb" ) ) == NULL )
			return -1 ;
	}

	return 0 ;
}

int close_files( ps )
	PARAMS ps ;
{
	if ( ps.sw_u == 1 ) fclose( ps.fpu ) ;
	if ( ps.sw_v == 1 ) fclose( ps.fpv ) ;

	return 0 ;
}

int write_img( ps, ds )
	PARAMS ps ;
	DIMS ds ;
{
	int y, n ;

	if ( ps.sw_u == 1 )
	for ( n=0; n<ps.ngrp; ++n )
	for ( y=0; y<ps.sy; ++y )
		fwrite( ds.ou[n][y], sizeof ***ds.ou, ps.sx, ps.fpu ) ;

	if ( ps.sw_v == 1 )
	for ( n=0; n<ps.ngrp; ++n )
	for ( y=0; y<ps.sy; ++y )
		fwrite( ds.ov[n][y], sizeof ***ds.ov, ps.sx, ps.fpv ) ;

	return 0 ;
}

int put_msg( ps )
	PARAMS ps ;
{
	printf( "PARAMS: \n parameter file; %s, result file; %s \n",
			ps.para_file, ps.result_file ) ;
	printf( " dx=%d, dy=%d, dt=%d\n", ps.dx, ps.dy, ps.dt ) ;
	printf( " sx=%d, sy=%d, st=%d\n", ps.sx, ps.sy, ps.st ) ;
	printf( " lt=%d, tintv=%d\n", ps.lt, ps.tintv ) ;
	printf( " Du=%e, Dv=%e\n", ps.Du, ps.Dv ) ;
	printf( " a0=%f, a1=%f, a2=%f, a3=%f\n", ps.a0, ps.a1, ps.a2, ps.a3 ) ;
	printf( " b=%f, eps=%e, mi=%f\n", ps.b, ps.eps, ps.mi ) ;
	printf( " delta_v=%f, phi_v=%f\n", ps.delta_v, ps.phi_v ) ;

	printf( " dt/dx^2 = %e, dt/dy^2=%e\n",
			(double)(ps.dx*ps.dx)/(double)ps.dt,
			(double)(ps.dy*ps.dy)/(double)ps.dt ) ;
	printf( " loop: lmin=%d, lmax=%d, delta=%e\n CN: r=%f\n",
			ps.lmin, ps.lmax, ps.delta, ps.r ) ;
	printf( " File name of a correlation map: iu=%s\n", ps.fnu ) ;
	printf( " Switch: sw_u=%d, sw_v=%d\n", ps.sw_u, ps.sw_v ) ;

	printf( " disparities (ngrp=%d):\n", ps.ngrp ) ;
	while ( ps.ngrp-- > 0 )
		printf( "\tdp[%d]=%f\n", ps.ngrp, ps.dp[ps.ngrp] ) ;

	#ifdef ZENSHIN-SABUN
		puts( "zenshin-sabun" ) ;
	#else
		puts( "chuushin-sabun" ) ;
	#endif

	fflush( stdout ) ;

	return 0 ;
}

int solve_u( ps, n, rCx, rCy, bij )
	PARAMS ps ;
	double **n, rCx, rCy, **bij ;
{
	int xl, x, xh, yl, y, yh, l ;
	double dif, A0, A1, t0 ;

	dif = ps.delta + 1. ;
	for ( l=0; dif>ps.delta && l<ps.lmax; ++l ) {
		dif = .0 ;

		/* xl=0; x=0; xh=1; yl=0; y=0; yh=1; */
		A0 = 1.+rCx+rCy ;
		A1 = rCx*n[0][1]+rCy*n[1][0]+bij[0][0] ;
		t0 = n[0][0] ;
		n[0][0] = A1/A0 ;
		dif += fabs(n[0][0]-t0) ;

		/* yl=0; y=0; yh=1; */
		A0 = 1.+2.*rCx+rCy ;
		for ( xl=0, x=1, xh=2; x<ps.sx-1; ++xl, ++x, ++xh ) {
			A1 = rCx*(n[0][xh]+n[0][xl])+rCy*n[1][x]+bij[0][x] ;
			t0 = n[0][x] ;
			n[0][x] = A1/A0 ;
			dif += fabs(n[0][x]-t0) ;
		}

		xl=ps.sx-2; x=ps.sx-1; xh=ps.sx-1; /* yl=0; y=0; yh=1;*/
		A0 = 1.+rCx+rCy ;
		A1 = rCx*n[0][xl]+rCy*n[1][x]+bij[0][x] ;
		t0 = n[0][x] ;
		n[0][x] = A1/A0 ;
		dif += fabs(n[0][x]-t0) ;

		for ( yl=0, y=1, yh=2; y<ps.sy-1; ++yl, ++y, ++yh ) {
			/* xl=0; x=0; xh=1 ; */
			A0 = 1.+rCx+2.*rCy ;
			A1 = rCx*n[y][xh]+rCy*(n[yh][x]+n[yl][x])+bij[y][0] ;
			t0 = n[y][0] ;
			n[y][0] = A1/A0 ;
			dif += fabs(n[y][0]-t0) ;

			A0 = 1.+2.*(rCx+rCy) ;
			for ( xl=0, x=1, xh=2; x<ps.sx-1; ++xl, ++x, ++xh ) {
				A1 = rCx*(n[y][xh]+n[y][xl])
					+ rCy*(n[yh][x]+n[yl][x])
					+ bij[y][x] ;
				t0 = n[y][x] ;
				n[y][x] = A1/A0 ;
				dif += fabs(n[y][x]-t0) ;
			}

			xl=ps.sx-2; x=ps.sx-1;
			A0 = 1.+rCx+2.*rCy ;
			A1 = rCx*n[y][xl]+rCy*(n[yh][x]+n[yl][x])+bij[y][x] ;
			t0 = n[y][x] ;
			n[y][x] = A1/A0 ;
			dif += fabs(n[y][x]-t0) ;
		}

		/* xl=0; x=0; xh=1; */
		yl=ps.sy-2; y=ps.sy-1; yh=ps.sy-1;

		A0 = 1.+rCx+rCy ;
		A1 = rCx*n[y][1]+rCy*n[yl][0]+bij[y][0] ;
		t0 = n[y][0] ;
		n[y][0] = A1/A0 ;
		dif += fabs(n[y][0]-t0) ;

		A0 = 1.+2.*rCx+rCy ;
		for ( xl=0, x=1, xh=2; x<ps.sx-1; ++xl, ++x, ++xh ) {
			A1 = rCx*(n[y][xh]+n[y][xl])+rCy*n[yl][x]+bij[y][x] ;
			t0 = n[y][x] ;
			n[y][x] = A1/A0 ;
			dif += fabs(n[y][x]-t0) ;
		}

		xl=ps.sx-2; x=ps.sx-1;
		A0 = 1.+rCx+rCy ;
		A1 = rCx*n[y][xl]+rCy*n[yl][x]+bij[y][x] ;
		t0 = n[y][x] ;
		n[y][x] = A1/A0 ;
		dif += fabs(n[y][x]-t0) ;
	}

	return l ;
}

int solve_v( ps, n, d, rCx, rCy, adx, ady, bij )
	PARAMS ps ;
	double **n, **d, rCx, rCy, **adx, **ady, **bij ;
{
	int xl, x, xh, yl, y, yh, l ;
	double dif, A0, A1, A2, A3, A4, t0 ;

	dif = ps.delta + 1. ;
	for ( l=0; dif>ps.delta && l<ps.lmax; ++l ) {
		dif = .0 ;

		/* x=0; xh=1; y=0; yh=1; */
		A0 = 1. + rCx*adx[0][0] + rCy*ady[0][0] ;
		A1 = rCx*adx[0][0]*n[0][1] ;
		A3 = rCy*ady[0][0]*n[1][0] ;
		t0 = n[0][0] ;
		n[0][0] = (A1+A3+bij[0][0])/A0 ;
		dif += fabs(n[0][0]-t0) ;

		/* y=0; yh=1; */
		for ( xl=0, x=1, xh=2; x<ps.sx-1; ++xl, ++x, ++xh ) {
			A0 = 1. + rCx*(adx[0][x]+adx[0][xl]) + rCy*ady[0][x] ;
			A1 = rCx*adx[0][x]*n[0][xh] ;
			A2 = rCx*adx[0][xl]*n[0][xl] ;
			A3 = rCy*ady[0][x]*n[1][x] ;
			t0 = n[0][x] ;
			n[0][x] = (A1+A2+A3+bij[0][x])/A0 ;
			dif += fabs(n[0][x]-t0) ;
		}

		xl=ps.sx-2; x=ps.sx-1; /* y=0; yh=1;*/
		A0 = 1. + rCx*adx[0][xl] + rCy*ady[0][x] ;
		A2 = rCx*adx[0][xl]*n[0][xl] ;
		A3 = rCy*ady[0][x]*n[1][x] ;
		t0 = n[0][x] ;
		n[0][x] = (A2+A3+bij[0][x])/A0 ;
		dif += fabs(n[0][x]-t0) ;

		for ( yl=0, y=1, yh=2; y<ps.sy-1; ++yl, ++y, ++yh ) {
			/* x=0; xh=1 ; */
			A0 = 1.+rCx*adx[y][0]+rCy*(ady[y][0]+ady[yl][0]) ;
			A1 = rCx*adx[y][0]*n[y][1] ;
			A3 = rCy*ady[y][0]*n[yh][0] ;
			A4 = rCy*ady[yl][0]*n[yl][0] ;
			t0 = n[y][0] ;
			n[y][0] = (A1+A3+A4+bij[y][0])/A0 ;
			dif += fabs(n[y][0]-t0) ;

			for ( xl=0, x=1, xh=2; x<ps.sx-1; ++xl, ++x, ++xh ) {
				A0 = 1. + rCx*(adx[y][x]+adx[y][xl])
					+ rCy*(ady[y][x]+ady[yl][x]) ;
				A1 = rCx*adx[y][x]*n[y][xh] ;
				A2 = rCx*adx[y][xl]*n[y][xl] ;
				A3 = rCy*ady[y][x]*n[yh][x] ;
				A4 = rCy*ady[yl][x]*n[yl][x] ;
				t0 = n[y][x] ;
				n[y][x] = (A1+A2+A3+A4+bij[y][x])/A0 ;
				dif += fabs(n[y][x]-t0) ;
			}

			xl=ps.sx-2; x=ps.sx-1;
			A0 = 1. + rCx*adx[y][xl]
				+ rCy*(ady[y][x]+ady[yl][x]) ;
			A2 = rCx*adx[y][xl]*n[y][xl] ;
			A3 = rCy*ady[y][x]*n[yh][x] ;
			A4 = rCy*ady[yl][x]*n[yl][x] ;
			t0 = n[y][x] ;
			n[y][x] = (A2+A3+A4+bij[y][x])/A0 ;
			dif += fabs(n[y][x]-t0) ;
		}

		/* x=0; xh=1; */
		yl=ps.sy-2; y=ps.sy-1;

		A0 = 1. + rCx*adx[y][0] + rCy*ady[yl][0] ;
		A1 = rCx*adx[y][0]*n[y][1] ;
		A4 = rCy*ady[yl][0]*n[yl][0] ;
		t0 = n[y][0] ;
		n[y][0] = (A1+A4+bij[y][0])/A0 ;
		dif += fabs(n[y][0]-t0) ;

		for ( xl=0, x=1, xh=2; x<ps.sx-1; ++xl, ++x, ++xh ) {
			A0 = 1. + rCx*(adx[y][x]+adx[y][xl])
				+ rCy*ady[yl][x] ;
			A1 = rCx*adx[y][x]*n[y][xh] ;
			A2 = rCx*adx[y][xl]*n[y][xl] ;
			A4 = rCy*ady[yl][x]*n[yl][x] ;
			t0 = n[y][x] ;
			n[y][x] = (A1+A2+A4+bij[y][x])/A0 ;
			dif += fabs(n[y][x]-t0) ;
		}

		xl=ps.sx-2; x=ps.sx-1;
		A0 = 1. + rCx*adx[y][xl] + rCy*ady[yl][x] ;
		A2 = rCx*adx[y][xl]*n[y][xl] ;
		A4 = rCy*ady[yl][x]*n[yl][x] ;
		t0 = n[y][x] ;
		n[y][x] = (A2+A4+bij[y][x])/A0 ;
		dif += fabs(n[y][x]-t0) ;
	}

	return l ;
}

int mk_a( ps, ou, max, adif, a, n0, dp )
	PARAMS ps ;
	int n0 ;
	double ***ou, **max, **adif, **a, *dp ;
{
	int n, x, y ;

	for ( y=0; y<ps.sy; ++y )
	for ( x=0; x<ps.sx; ++x ) {
		max[y][x] = -1. ;
		adif[y][x] = -1. ;
	}

	for ( n=0; n<ps.ngrp; ++n )
		if ( n!=n0 )
		for ( y=0; y<ps.sy; ++y )
		for ( x=0; x<ps.sx; ++x ) {
			if ( ou[n][y][x]>max[y][x] ) {
				max[y][x] = ou[n][y][x] ;
				adif[y][x] = fabs( dp[n0]-dp[n] ) ;
			}

			if ( (x+n0-n)>=0 && (x+n0-n)<ps.sx )
			if ( ou[n][y][x+n0-n]>max[y][x+n0-n] ) {
				max[y][x+n0-n] = ou[n][y][x+n0-n] ;
				adif[y][x+n0-n] = fabs( dp[n0]-dp[n] ) ;
			}
		}

	for ( y=0; y<ps.sy; ++y )
	for ( x=0; x<ps.sx; ++x )
		a[y][x] = ps.a0 + max[y][x]*(1.+tanh(adif[y][x]-ps.a1))/2. ;

	return 0 ;
}


int mk_bu( ps, ou, ov, o, bij, rCx, rCy, p0, p1, p2 )
	PARAMS ps ;
	double **ou, **ov, **o, **bij ;
	double rCx, rCy, p0, **p1, **p2 ;
{
	int x, y, xl, xh, yl, yh ;
	double e, px, py, pz, pr ;

	e = 1./(double)ps.dt ;

	/* x=0, y=0 */
	px = rCx*(o[0][1]+o[0][0]) ;
	py = rCy*(o[1][0]+o[0][0]) ;
	pz = (1.-2.*(rCx+rCy))*o[y][x] ;
	pr = e*fuv(p0,p1[0][0],p2[0][0],ou[0][0],ov[0][0]) ;
	bij[0][0] = px + py + pz + pr ;

	/* y = 0 ; yh = 1 ; */
	for ( xl=0, x=1, xh=2; x<ps.sx-1; ++xl, ++x, ++xh ) {
		px = rCx*(o[0][xh]+o[0][xl]) ;
		py = rCy*(o[1][x]+o[0][x]) ;
		pz = (1.-2.*(rCx+rCy))*o[0][x] ;
		pr = e*fuv(p0,p1[0][x],p2[0][x],ou[0][x],ov[0][x]) ;
		bij[0][x] = px + py + pz + pr ;
	}

	x = ps.sx-1 ; xl = x-1 ;
	/* y=0 */
	px = rCx*(o[0][x]+o[0][xl]) ;
	py = rCy*(o[1][x]+o[0][x]) ;
	pz = (1.-2.*(rCx+rCy))*o[y][x] ;
	pr = e*fuv(p0,p1[0][x],p2[0][x],ou[0][x],ov[0][x]) ;
	bij[0][x] = px + py + pz + pr ;

	for ( yl=0, y=1, yh=2; y<ps.sy-1; ++yl, ++y, ++yh ) {
		/* x=0 */
		px = rCx*(o[y][1]+o[y][0]) ;
		py = rCy*(o[yh][0]+o[yl][0]) ;
		pz = (1.-2.*(rCx+rCy))*o[y][x] ;
		pr = e*fuv(p0,p1[y][0],p2[y][0],ou[y][0],ov[y][0]) ;
		bij[y][0] = px + py + pz + pr ;

		for ( xl=0, x=1, xh=2; x<ps.sx-1; ++xl, ++x, ++xh ) {
			px = rCx*(o[y][xh]+o[y][xl]) ;
			py = rCy*(o[yh][x]+o[yl][x]) ;
			pz = (1.-2.*(rCx + rCy))*o[y][x] ;
			pr = e*fuv(p0,p1[y][x],p2[y][x],ou[y][x],ov[y][x]) ;
			bij[y][x] = px + py + pz + pr ;
		}

		x = ps.sx-1 ;
		xl = x-1 ;
		px = rCx*(o[y][x]+o[y][xl]) ;
		py = rCy*(o[yh][x]+o[yl][x]) ;
		pz = (1.-2.*(rCx+rCy))*o[y][x] ;
		pr = e*fuv(p0,p1[y][x],p2[y][x],ou[y][x],ov[y][x]) ;
		bij[y][x] = px + py + pz + pr ;
	}

	/* x=0: at the bottom left corner */
	y = ps.sy-1 ; yl = y-1 ; 
	px = rCx*(o[y][1]+o[y][0]) ;
	py = rCy*(o[y][0]+o[yl][0]) ;
	pz = (1.-2.*(rCx+rCy))*o[y][0] ;
	pr = e*fuv(p0,p1[y][0],p2[y][0],ou[y][0],ov[y][0]) ;
	bij[y][0] = px + py + pz + pr ;

	y = ps.sy-1 ; yl = y-1 ;
	for ( xl=0, x=1, xh=2; x<ps.sx-1; ++xl, ++x, ++xh ) {
		px = rCx*(o[y][xh]+o[y][xl]) ;
		py = rCy*(o[y][x]+o[yl][x]) ;
		pz = (1.-2.*(rCx+rCy))*o[y][x] ;
		pr = e*fuv(p0,p1[y][x],p2[y][x],ou[y][x],ov[y][x]) ;
		bij[y][x] = px + py + pz + pr ;
	}

	/* at the bottom right corner */
	x = ps.sx-1 ; xl = x-1 ;
	y = ps.sy-1 ; yl = y-1 ;
	px = rCx*(o[y][x]+o[y][xl]) ;
	py = rCy*(o[y][x]+o[yl][x]) ;
	pz = (1.-2.*(rCx+rCy))*o[y][x] ;
	pr = e*fuv(p0,p1[y][x],p2[y][x],ou[y][x],ov[y][x]) ;
	bij[y][x] = px + py + pz + pr ;

	return 0 ;
}

int mk_bv( ps, ou, ov, o, d, adx, ady, bij, rCx, rCy, p0 )
	PARAMS ps ;
	double **ou, **ov, **o, **d, **adx, **ady, **bij ;
	double rCx, rCy, p0 ;
{
	int x, y, xl, xh, yl, yh ;
	double e, px, py, pz, pr ;

	e = 1./(double)ps.dt ;

	/* x=0, y=0 */
	px = rCx*(adx[0][0]*o[0][1]+d[0][0]*o[0][0]) ;
	py = rCy*(ady[0][0]*o[1][0]+d[0][0]*o[0][0]) ;
	pz = rCx*(adx[0][0]+d[0][0])+rCy*(ady[0][0]+d[0][0]) ;
	pr = e*guv(p0,ou[0][0],ov[0][0]) ;
	bij[0][0] = px + py + (1.-pz)*o[0][0] + pr ;

	/* y = 0 ; yh = 1 ; */
	for ( xl=0, x=1, xh=2; x<ps.sx-1; ++xl, ++x, ++xh ) {
		px = rCx*(adx[0][x]*o[0][xh]+adx[0][xl]*o[0][xl]) ;
		py = rCy*(ady[0][x]*o[1][x]+d[0][x]*o[0][x]) ;
		pz = rCx*(adx[0][x]+adx[0][xl])+rCy*(ady[0][x]+d[0][x]) ;
		pr = e*guv(p0,ou[0][x],ov[0][x]) ;
		bij[0][x] = px + py + (1.-pz)*o[0][x] + pr ;
	}

	x = ps.sx-1 ; xl = x-1 ;
	/* y=0 */
	px = rCx*(d[0][x]*o[0][x]+adx[0][xl]*o[0][xl]) ;
	py = rCy*(ady[0][x]*o[1][x]+d[0][x]*o[0][x]) ;
	pz = rCx*(d[0][x]+adx[0][xl])+rCy*(ady[0][x]+d[0][x]) ;
	pr = e*guv(p0,ou[0][x],ov[0][x]) ;
	bij[0][x] = px + py + (1.-pz)*o[0][x] + pr ;

	for ( yl=0, y=1, yh=2; y<ps.sy-1; ++yl, ++y, ++yh ) {
		/* x=0 */
		px = rCx*(adx[y][0]*o[y][1]+d[y][0]*o[y][0]) ;
		py = rCy*(ady[y][0]*o[yh][0]+ady[yl][0]*o[yl][0]) ;
		pz = rCx*(adx[y][0]+d[y][0])+rCy*(ady[y][0]+ady[yl][0]) ;
		pr = e*guv(p0,ou[y][0],ov[y][0]) ;
		bij[y][0] = px + py + (1.-pz)*o[y][0] + pr ;

		for ( xl=0, x=1, xh=2; x<ps.sx-1; ++xl, ++x, ++xh ) {
			px = rCx*(adx[y][x]*o[y][xh]+adx[y][xl]*o[y][xl]) ;
			py = rCy*(ady[y][x]*o[yh][x]+ady[yl][x]*o[yl][x]) ;
			pz = rCx*(adx[y][x]+adx[y][xl])
			   + rCy*(ady[y][x]+ady[yl][x]) ;
			pr = e*guv(p0,ou[y][x],ov[y][x]) ;
			bij[y][x] = px + py + (1.-pz)*o[y][x] + pr ;
		}

		x = ps.sx-1 ;
		xl = x-1 ;
		px = rCx*(d[y][x]*o[y][x]+adx[y][xl]*o[y][xl]) ;
		py = rCy*(ady[y][x]*o[yh][x]+ady[yl][x]*o[yl][x]) ;
		pz = rCx*(d[y][x]+adx[y][xl])+rCy*(ady[y][x]+ady[yl][x]) ;
		pr = e*guv(p0,ou[y][x],ov[y][x]) ;
		bij[y][x] = px + py + (1.-pz)*o[y][x] + pr ;
	}

	/* x=0: at the bottom left corner */
	y = ps.sy-1 ; yl = y-1 ; 
	px = rCx*(adx[y][0]*o[y][1]+d[y][0]*o[y][0]) ;
	py = rCy*(d[y][0]*o[y][0]+ady[yl][0]*o[yl][0]) ;
	pz = rCx*(adx[y][0]+d[y][0])+rCy*(d[y][0]+ady[yl][0]) ;
	pr = e*guv(p0,ou[y][0],ov[y][0]) ;
	bij[y][0] = px + py + (1.-pz)*o[y][0] + pr ;

	y = ps.sy-1 ; yl = y-1 ;
	for ( xl=0, x=1, xh=2; x<ps.sx-1; ++xl, ++x, ++xh ) {
		px = rCx*(adx[y][x]*o[y][xh]+adx[y][xl]*o[y][xl]) ;
		py = rCy*(d[y][x]*o[y][x]+ady[yl][x]*o[yl][x]) ;
		pz = rCx*(adx[y][x]+adx[y][xl])+rCy*(d[y][x]+ady[yl][x]) ;
		pr = e*guv(p0,ou[y][x],ov[y][x]) ;
		bij[y][x] = px + py + (1.-pz)*o[y][x] + pr ;
	}

	/* at the bottom right corner */
	x = ps.sx-1 ; xl = x-1 ;
	y = ps.sy-1 ; yl = y-1 ;
	px = rCx*(d[y][x]*o[y][x]+adx[y][xl]*o[y][xl]) ;
	py = rCy*(d[y][x]*o[y][x]+ady[yl][x]*o[yl][x]) ;
	pz = rCx*(d[y][x]+adx[y][xl])+rCy*(d[y][x]+ady[yl][x]) ;
	pr = e*guv(p0,ou[y][x],ov[y][x]) ;
	bij[y][x] = px + py + (1.-pz)*o[y][x] + pr ;

	return 0 ;
}

int set_initial_state( ps, ds )
	PARAMS ps ;
	DIMS ds ;
{
	int x, y, n, s ;
	FILE *fp ;
	double *p, tmp ;
	char str[STR_LEN] ;

	if ( (fp=fopen(ps.fnu,"rb")) == NULL ) {
		fprintf( stderr, "error: cannot open file %s.\n", ps.fnu ) ;
		return -1 ;
	} else {
		for ( n=0; n<ps.ngrp; ++n )
		for ( y=0; y<ps.sy; ++y ) {
			fread( ds.iu[n][y], sizeof ***ds.iu, ps.sx, fp ) ;
			for ( x=0; x<ps.sx; ++x )
				ds.iu[n][y][x] *= ps.mi ;
		}

		fclose( fp ) ;
	}

	for ( n=0; n<ps.ngrp; ++n )
	for ( y=0; y<ps.sy; ++y )
	for ( x=0; x<ps.sx; ++x ) {
		ds.nu[n][y][x] = ds.ou[n][y][x] = 0. ;
		ds.nv[n][y][x] = ds.ov[n][y][x] = 0. ;
	}

	return 0 ;
}

int av_d_xy( ps, d, av_d_x, av_d_y )
	PARAMS ps ;
	double **d, **av_d_x, **av_d_y ;
{
	int x, y ;

	for ( y=0; y<ps.sy-1; ++y )
	for ( x=0; x<ps.sx-1; ++x ) {
		av_d_x[y][x] = (d[y][x+1]+d[y][x])/2. ;
		av_d_y[y][x] = (d[y+1][x]+d[y][x])/2. ;
	}

	y = ps.sy-1 ;
	for ( x=0; x<ps.sx-1; ++x )
		av_d_x[y][x] = (d[y][x+1]+d[y][x])/2. ;
	x = ps.sx-1 ;
	for ( y=0; y<ps.sy; ++y )
		av_d_x[y][x] = d[y][x] ;

	x = ps.sx-1 ;
	for ( y=0; y<ps.sy-1; ++y )
		av_d_y[y][x] = (d[y+1][x]+d[y][x])/2. ;
	y = ps.sy-1 ;
	for ( x=0; x<ps.sx; ++x )
		av_d_y[y][x] = d[y][x] ;

	return 0 ;
}

double get_dv_sub( ps, vx, vy )
	PARAMS ps ;
	double vx, vy ;
{
	if ( vx==0. && vy==0. )
		return 1. ;
	else {
		double theeta_v, tmp ;

		theeta_v = atan2( vy, vx ) ;
		tmp = 1. - ps.delta_v * cos( 2.*(theeta_v-ps.phi_v) ) ;

		return ( 1. / sqrt( tmp ) ) ;
	}
}

int get_dv( ps, ov, dv )
	PARAMS ps ;
	double **ov, **dv ;
{
	int x, y ;
	double vx, vy ;

	/* y=0 and x=0 */
	vx = ov[0][1]-ov[0][0] ;
	vy = ov[1][0]-ov[0][0] ;
	dv[0][0] = get_dv_sub( ps, vx, vy ) ;

	/* y=0 */
	for ( x=1; x<ps.sx-1; ++x ) {
		vx = (ov[0][x+1]-ov[0][x-1])/2. ;
		vy = ov[1][x]-ov[0][x] ;
		dv[0][x] = get_dv_sub( ps, vx, vy ) ;
	}

	/* y=0, x=ps.sx-1 */
	vx = ov[0][ps.sx-1]-ov[0][ps.sx-2] ;
	vy = ov[1][ps.sx-1]-ov[0][ps.sx-1] ;
	dv[0][ps.sx-1] = get_dv_sub( ps, vx, vy ) ;

	for ( y=1; y<ps.sy-1; ++y ) {
		/* x=0 */
		vx = ov[y][1]-ov[y][0] ;
		vy = (ov[y+1][0]-ov[y-1][0])/2. ;
		dv[y][0] = get_dv_sub( ps, vx, vy ) ;

		for ( x=1; x<ps.sx-1; ++x ) {
			vx = (ov[y][x+1]-ov[y][x-1])/2. ;
			vy = (ov[y+1][x]-ov[y-1][x])/2. ;
			dv[y][x] = get_dv_sub( ps, vx, vy ) ;
		}

		/* x=ps.sx-1 */
		vx = ov[y][ps.sx-1]-ov[y][ps.sx-2] ;
		vy = (ov[y+1][ps.sx-1]-ov[y-1][ps.sx-1])/2. ;
		dv[y][ps.sx-1] = get_dv_sub( ps, vx, vy ) ;
	}

	/* x=0, y=ps.sy-1 */
	vx = ov[ps.sy-1][1]-ov[ps.sy-1][0] ;
	vy = ov[ps.sy-1][0]-ov[ps.sy-2][0] ;
	dv[ps.sy-1][0] = get_dv_sub( ps, vx, vy ) ;

	/* y=ps.sy-1 */
	for ( x=1; x<ps.sx-1; ++x ) {
		vx = (ov[ps.sy-1][x+1]-ov[ps.sy-1][x-1])/2. ;
		vy = ov[ps.sy-1][x]-ov[ps.sy-2][x] ;
		dv[ps.sy-1][x] = get_dv_sub( ps, vx, vy ) ;
	}

	/* x=ps.sx-1, y=ps.sy-1 */
	vx = ov[ps.sy-1][ps.sx-1]-ov[ps.sy-1][ps.sx-2] ;
	vy = ov[ps.sy-1][ps.sx-1]-ov[ps.sy-2][ps.sx-1] ;
	dv[ps.sy-1][ps.sx-1] = get_dv_sub( ps, vx, vy ) ;

	return 0 ;
}

#ifdef ZENSHIN-SABUN
int get_dv( ps, ov, dv )
	PARAMS ps ;
	double **ov, **dv ;
{
	int x, y ;
	double vx, vy ;

	/* y=0 and x=0 */
	vx = ov[0][1]-ov[0][0] ;
	vy = ov[1][0]-ov[0][0] ;
	dv[0][0] = get_dv_sub( ps, vx, vy ) ;

	/* y=0 */
	for ( x=1; x<ps.sx-1; ++x ) {
		vx = ov[0][x+1]-ov[0][x] ;
		vy = ov[1][x]-ov[0][x] ;
		dv[0][x] = get_dv_sub( ps, vx, vy ) ;
	}

	/* y=0, x=ps.sx-1 */
	vx = ov[0][ps.sx-1]-ov[0][ps.sx-2] ;
	vy = ov[1][ps.sx-1]-ov[0][ps.sx-1] ;
	dv[0][ps.sx-1] = get_dv_sub( ps, vx, vy ) ;

	for ( y=1; y<ps.sy-1; ++y ) {
		/* x=0 */
		vx = ov[y][1]-ov[y][0] ;
		vy = ov[y+1][0]-ov[y][0] ;
		dv[y][0] = get_dv_sub( ps, vx, vy ) ;

		for ( x=1; x<ps.sx-1; ++x ) {
			vx = ov[y][x+1]-ov[y][x] ;
			vy = ov[y+1][x]-ov[y][x] ;
			dv[y][x] = get_dv_sub( ps, vx, vy ) ;
		}

		/* x=ps.sx-1 */
		vx = ov[y][ps.sx-1]-ov[y][ps.sx-2] ;
		vy = ov[y+1][ps.sx-1]-ov[y][ps.sx-1] ;
		dv[y][ps.sx-1] = get_dv_sub( ps, vx, vy ) ;
	}

	/* x=0, y=ps.sy-1 */
	vx = ov[ps.sy-1][1]-ov[ps.sy-1][0] ;
	vy = ov[ps.sy-1][0]-ov[ps.sy-2][0] ;
	dv[ps.sy-1][0] = get_dv_sub( ps, vx, vy ) ;

	/* y=ps.sy-1 */
	for ( x=1; x<ps.sx-1; ++x ) {
		vx = ov[ps.sy-1][x+1]-ov[ps.sy-1][x] ;
		vy = ov[ps.sy-1][x]-ov[ps.sy-2][x] ;
		dv[ps.sy-1][x] = get_dv_sub( ps, vx, vy ) ;
	}

	/* x=ps.sx-1, y=ps.sy-1 */
	vx = ov[ps.sy-1][ps.sx-1]-ov[ps.sy-1][ps.sx-2] ;
	vy = ov[ps.sy-1][ps.sx-1]-ov[ps.sy-2][ps.sx-1] ;
	dv[ps.sy-1][ps.sx-1] = get_dv_sub( ps, vx, vy ) ;

	return 0 ;
}
#endif

void *sub( pthr )
	void *pthr ;
{
	PARAMS_PTH *p = (PARAMS_PTH *)pthr ;
	PARAMS ps = p->ps ;
	int n0 = p->n0 ;
	double **iu = p->iu ;
	double **ou=p->ou, **nu=p->nu, **ov=p->ov, **nv=p->nv ;
	double **dv=p->dv, **av_dv_x=p->av_dv_x, **av_dv_y=p->av_dv_y ;
	double **a=p->a, **bij=p->bij, **max=p->max, **adif=p->adif ;
	double ***u_all=p->u_all ;

	while ( 1 ) {
		sema_wait( &(p->start) ) ;

		get_dv( ps, ov, dv ) ;
		av_d_xy( ps, dv, av_dv_x, av_dv_y ) ;

		mk_a (	ps, u_all, max, adif, a, n0, ps.dp ) ;
		mk_bu (	ps, ou, ov, ou, bij,
			ps.r1Cux, ps.r1Cuy, ps.eps, a, iu ) ;
		solve_u(ps, nu, ps.r0Cux, ps.r0Cuy, bij ) ;

		mk_bv (	ps, ou, ov, ov, dv, av_dv_x, av_dv_y, bij,
			ps.r1Cvx, ps.r1Cvy, ps.b ) ;
		solve_v(ps, nv, dv, ps.r0Cvx, ps.r0Cvy,
			av_dv_x, av_dv_y, bij ) ;

		sema_post( &(p->stop) ) ;
	}
}


int mk_b( ps, o, bij, rCx, rCy )
	PARAMS ps ;
	double **o, **bij ;
	double rCx, rCy ;
{
	int x, y, xl, xh, yl, yh ;
	double px, py, pz ;

	/* x=0, y=0 */
	px = rCx*(o[0][1]+o[0][0]) ;
	py = rCy*(o[1][0]+o[0][0]) ;
	pz = (1.-2.*(rCx+rCy))*o[y][x] ;
	bij[0][0] = px + py + pz ;

	/* y = 0 ; yh = 1 ; */
	for ( xl=0, x=1, xh=2; x<ps.sx-1; ++xl, ++x, ++xh ) {
		px = rCx*(o[0][xh]+o[0][xl]) ;
		py = rCy*(o[1][x]+o[0][x]) ;
		pz = (1.-2.*(rCx+rCy))*o[0][x] ;
		bij[0][x] = px + py + pz ;
	}

	x = ps.sx-1 ; xl = x-1 ;
	/* y=0 */
	px = rCx*(o[0][x]+o[0][xl]) ;
	py = rCy*(o[1][x]+o[0][x]) ;
	pz = (1.-2.*(rCx+rCy))*o[y][x] ;
	bij[0][x] = px + py + pz ;

	for ( yl=0, y=1, yh=2; y<ps.sy-1; ++yl, ++y, ++yh ) {
		/* x=0 */
		px = rCx*(o[y][1]+o[y][0]) ;
		py = rCy*(o[yh][0]+o[yl][0]) ;
		pz = (1.-2.*(rCx+rCy))*o[y][x] ;
		bij[y][0] = px + py + pz ;

		for ( xl=0, x=1, xh=2; x<ps.sx-1; ++xl, ++x, ++xh ) {
			px = rCx*(o[y][xh]+o[y][xl]) ;
			py = rCy*(o[yh][x]+o[yl][x]) ;
			pz = (1.-2.*(rCx + rCy))*o[y][x] ;
			bij[y][x] = px + py + pz ;
		}

		x = ps.sx-1 ;
		xl = x-1 ;
		px = rCx*(o[y][x]+o[y][xl]) ;
		py = rCy*(o[yh][x]+o[yl][x]) ;
		pz = (1.-2.*(rCx+rCy))*o[y][x] ;
		bij[y][x] = px + py + pz ;
	}

	/* x=0: at the bottom left corner */
	y = ps.sy-1 ; yl = y-1 ; 
	px = rCx*(o[y][1]+o[y][0]) ;
	py = rCy*(o[y][0]+o[yl][0]) ;
	pz = (1.-2.*(rCx+rCy))*o[y][0] ;
	bij[y][0] = px + py + pz ;

	y = ps.sy-1 ; yl = y-1 ;
	for ( xl=0, x=1, xh=2; x<ps.sx-1; ++xl, ++x, ++xh ) {
		px = rCx*(o[y][xh]+o[y][xl]) ;
		py = rCy*(o[y][x]+o[yl][x]) ;
		pz = (1.-2.*(rCx+rCy))*o[y][x] ;
		bij[y][x] = px + py + pz ;
	}

	/* at the bottom right corner */
	x = ps.sx-1 ; xl = x-1 ;
	y = ps.sy-1 ; yl = y-1 ;
	px = rCx*(o[y][x]+o[y][xl]) ;
	py = rCy*(o[y][x]+o[yl][x]) ;
	pz = (1.-2.*(rCx+rCy))*o[y][x] ;
	bij[y][x] = px + py + pz ;

	return 0 ;
}


int calc( ps, ds )
	PARAMS ps ;
	DIMS ds ;
{
	int x, y, n, t ;
	PARAMS_PTH *thr ;
	double dudt, max_dudt ;

	if ( put_msg( ps ) == -1 )
		return -1 ;
	else if ( open_files( &ps ) == -1 )
		return -1 ;
	else if ( set_initial_state( ps, ds ) == -1 )
		return -1 ;
	else if ( (thr=(PARAMS_PTH *)malloc(ps.ngrp*sizeof(*thr)))==NULL )
		return -1 ;
	else
		write_img( ps, ds ) ;

        for ( n=0; n<ps.ngrp; ++n ) {
		thr[n].ps = ps ;
		thr[n].n0 = n ;

		thr[n].iu = ds.iu[n]  ;

		thr[n].ou = ds.ou[n] ;
		thr[n].nu = ds.nu[n] ;
		thr[n].ov = ds.ov[n] ;
		thr[n].nv = ds.nv[n] ;

		thr[n].dv = ds.dv[n] ;

		thr[n].av_dv_x = ds.av_dv_x[n] ;
		thr[n].av_dv_y = ds.av_dv_y[n] ;

		thr[n].a = ds.a[n]  ;
		thr[n].bij = ds.bij[n] ;
		thr[n].max = ds.max[n]  ;
		thr[n].adif = ds.adif[n]  ;
		thr[n].u_all = ds.ou ;

		sema_init( &(thr[n].start), 0, USYNC_THREAD, NULL ) ;
		sema_init( &(thr[n].stop ), 0, USYNC_THREAD, NULL ) ;

		t = thr_create( NULL, NULL,
			sub, &(thr[n]), THR_BOUND, &(thr[n].tid) ) ;
		if ( t ) {
			close_files( ps ) ;
			free( thr ) ;
			return -1 ;
		}
	}

	for ( t=1; t<ps.st; ++t ) {
		for ( n=0; n<ps.ngrp; ++n )
			sema_post( &(thr[n].start) ) ;

		if ( (t%ps.tintv)==0 ) {
			printf( "t=%f, u[%d][%d][%d]=%e\n",
				(double)t/(double)ps.dt,
				ps.ngrp/2, ps.sy/2, ps.sx/2,
				ds.ou[ps.ngrp/2][ps.sy/2][ps.sx/2] ) ;
			write_img( ps, ds ) ;
		}

		for ( n=0; n<ps.ngrp; ++n )
			sema_wait( &(thr[n].stop) ) ;

		if ( (t%ps.tintv)==0 || t<ps.dt ) {
			max_dudt = -1.0 ;
			for ( n=0; n<ps.ngrp; ++n )
			for ( y=0; y<ps.sy; ++y )
			for ( x=0; x<ps.sx; ++x ) {
				dudt = fabs( ds.nu[n][y][x]-ds.ou[n][y][x] )
					* ps.dt ;
				if ( dudt > max_dudt )
					max_dudt = dudt ;
			}
			printf( "t=%f, max_dudt=%e\n",
				(double)t/(double)ps.dt, max_dudt ) ;
			fflush( stdout ) ;
		}

		for ( n=0; n<ps.ngrp; ++n )
		for ( y=0; y<ps.sy; ++y )
		for ( x=0; x<ps.sx; ++x ) {
			ds.ou[n][y][x] = ds.nu[n][y][x] ;
			ds.ov[n][y][x] = ds.nv[n][y][x] ;
		}
	}

	write_img( ps, ds ) ;
	close_files( ps ) ;

	return 0 ;
}

