/*
	Stereo algorithm with anisotropic reaction-diffusion equations.
	~/vision/mngroup-tanh4.nud5.lwp.aid3
	previous:
		~/vision/mngroup-tanh4.nud5.lwp.aid2
		~/vision/mngroup-tanh4.nud5.lwp
		~/vision/mngroup-tanh4.nud5
		~/vision/mngroup-tanh4
		~/vision/ed.fn2d.nud.sor
*/
#define _REENTRANT
#include <thread.h>
#include <an.h>

#define STR_LEN	512

typedef struct PARAMS {
	/* number of groups */
	int ngrp ;

	/* files */
	char *para_file, *result_file ;

	/* discretisation and system size parameters */
	int dx, dy, dt ;
	int lt ;
	int sx, sy, st ;

	/* time interval for displaying the result */
	int tintv ;

	/* file name storing a correlation map */
	char fnu[STR_LEN] ;	/* correlation map => iu */

	/* reaction-diffusion parameters */
	double Du, Dv, a0, a1, a2, a3, b, eps ;
	double Cux, Cuy, Cvx, Cvy ;
	double r0Cux, r0Cuy, r0Cvx, r0Cvy ;
	double r1Cux, r1Cuy, r1Cvx, r1Cvy ;

	/* for anisotropic diffusion: dv */
	double delta_v, phi_v ;	/* delta_v: strength, phi_v: direction */

	/* for numerical computation */

	/* Crank-Nicolson */
	double r ;

	/* loop number for solving simultaneous equations */
	double delta ;
	int lmin, lmax ;

	/* on stimulus */
	double mi ;

	/* disparities */
	double *dp ;

	/* file pointers */
	FILE *fpu, *fpv ;

	/* switch on u and v for data save */
	int sw_u, sw_v ;
} PARAMS ;

typedef struct DIMS {
	/* provided as a source term at every iteration */
	double ***iu ;

	/* variables representing all of N groups (3D:x,y,n) */
	double ***ou, ***nu, ***ov, ***nv ;

	/* non-uniform diffusion coefficients (2D:x,y) */
	double ***dv ;

	/* average */
	double ***av_dv_x, ***av_dv_y ;

	double ***a ;
	double ***bij ;

	/* maximum value of u */
	double ***max, ***adif ;
} DIMS ;

typedef struct PARAMS_PTH {
	PARAMS ps ;
	int n0 ;

	double **iu ;

	double **ou, **nu, **ov, **nv ;

	double **dv, **av_dv_x, **av_dv_y ;

	double **a, **bij, **max, **adif ;
	double ***u_all ;

	thread_t tid ;
	sema_t start, stop ;
} PARAMS_PTH ;

