/*
	overlap_cor3:
		varies disparity continuously
			in the range of dp_min - dp_max.
	5 points correlation
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <an.h>

#define STR_LEN		512

int read_stereo_images( fn, il, ir, img )
	char *fn ;
	unsigned char **ir, **il ;
	AN_IMAGE_COORD img ;
{
	FILE *fp ;
	char str[STR_LEN] ;
	int y ;

	if ( (fp=fopen(strcat(strcpy(str,fn),".l"),"rb"))==NULL ) {
		return -1 ;
	} else {
		for ( y=0; y<img.y; ++y )
			fread( il[y], sizeof **il, img.x, fp ) ;

		fclose( fp ) ;
	}

	if ( (fp=fopen(strcat(strcpy(str,fn),".r"),"rb"))==NULL ) {
		return -1 ;
	} else {
		for ( y=0; y<img.y; ++y )
			fread( ir[y], sizeof **ir, img.x, fp ) ;

		fclose( fp ) ;
	}

	return 0 ;
}

int cor3( ir, il, rd, img, fp )
	unsigned char **ir, **il ;
	double **rd ;
	AN_IMAGE_COORD img ;
	FILE *fp ;
{
	int x, y, d, xr, xl, i, j, lxy ;
	double cor, s, sr, sl, avr, avl ;
	int dif, dp_min, dp_max ;	/* disparity */
	char str[STR_LEN] ;

	puts( "disparity: dp_min =?" ) ;
	dp_min = atoi( fgets( str, STR_LEN, stdin ) ) ;
	puts( "disparity: dp_max =?" ) ;
	dp_max = atoi( fgets( str, STR_LEN, stdin ) ) ;
	if ( dp_min > dp_max ) {
		dif = dp_max ;
		dp_max = dp_min ;
		dp_min = dif ;
	}

	printf( "#cor3: 5 points cross-correlation\n" ) ;
	printf( "#disparity: dp_min=%d, dp_max=%d\n", dp_min, dp_max ) ;

	puts( "#dp:" ) ;
	for ( dif=dp_min; dif<=dp_max; ++dif ) {
		printf( "%d\n", dif ) ;
		for ( y=0; y<img.y; ++y )
		for ( x=0; x<img.x; ++x )
			rd[y][x] = 0. ;

		for ( y=1; y<img.y-1; ++y )
		for ( x=1; x<img.x-1; ++x ) {
			xr = x+dif ;
			xl = x ;
			if ( xr>=1 && xr<img.x-1 && xl>=1 && xl<img.x-1 ) {

				sr  = ir[y][xr] ;
				sr += ir[y][xr-1] ;
				sr += ir[y][xr+1] ;
				sr += ir[y-1][xr] ;
				sr += ir[y+1][xr] ;
				avr = sr/5. ;

				sl  = il[y][xl] ;
				sl += il[y][xl-1] ;
				sl += il[y][xl+1] ;
				sl += il[y-1][xl] ;
				sl += il[y+1][xl] ;
				avl = sl/5. ;

				s   = (ir[y][xr]-avr)*(il[y][xl]-avl) ;
				s  += (ir[y][xr-1]-avr)*(il[y][xl-1]-avl) ;
				s  += (ir[y][xr+1]-avr)*(il[y][xl+1]-avl) ;
				s  += (ir[y-1][xr]-avr)*(il[y-1][xl]-avl) ;
				s  += (ir[y+1][xr]-avr)*(il[y+1][xl]-avl) ;

				sr  = pow( ir[y][xr]-avr, 2.) ;
				sr += pow( ir[y][xr-1]-avr, 2.) ;
				sr += pow( ir[y][xr+1]-avr, 2.) ;
				sr += pow( ir[y-1][xr]-avr, 2.) ;
				sr += pow( ir[y+1][xr]-avr, 2.) ;

				sl  = pow( il[y][xl]-avl, 2.) ;
				sl += pow( il[y][xl-1]-avl, 2.) ;
				sl += pow( il[y][xl+1]-avl, 2.) ;
				sl += pow( il[y-1][xl]-avl, 2.) ;
				sl += pow( il[y+1][xl]-avl, 2.) ;

				if ( sr > .0 || sl > .0 )
					cor = s / ( sr>sl ? sr: sl ) ;
				else
					cor = .0 ;

rd[y][x] = cor ;
/*
				if ( cor<0. )
					rd[y][x] = 0. ;
				else if ( cor<=1. )
					rd[y][x] = cor ;
				else
					rd[y][x] = 1. ;
*/
			}
		}

		for ( y=0; y<img.y; ++y )
			fwrite( rd[y], sizeof **rd, img.x, fp ) ;
	}

	return 0 ;
}

int main( argc, argv )
	int argc ;
	char **argv ;
{
	AN_IMAGE_COORD img ;
	FILE *fp ;
	unsigned char **ir, **il ;
	double **rd ;

	if ( argc!=3 ) {
		fprintf( stderr, "Usage: %s file file\n", argv[0] ) ;
		return 1 ;
	} else if ( an_get_img( &img ) == -1 ) {
		fputs( "Error: cannot find image environment.\n", stderr ) ;
		return 2 ;
	} else if ( (fp=fopen(argv[2],"wb"))==NULL ) {
		fprintf( stderr, "error: cannot open the correlation file %s\n",
			argv[2] ) ;
		return 2 ;
	} else if ( (il=(unsigned char **)an_malloc_2D(img.x,img.y,sizeof**il))
			==NULL ) {
		fputs( "error: cannot allocate memory.\n", stderr ) ;
		fclose( fp ) ;
		return 2 ;
	} else if ( (ir=(unsigned char **)an_malloc_2D(img.x,img.y,sizeof**ir))
			==NULL ) {
		fputs( "error: cannot allocate memory.\n", stderr ) ;
		free( il ) ;
		fclose( fp ) ;
		return 2 ;
	} else if ( read_stereo_images( argv[1], il, ir, img ) == -1 ) {
		fputs( "error: cannot read stereo images.\n", stderr ) ;
		free( il ) ;
		free( ir ) ;
		fclose( fp ) ;
		return 2 ;
	} else if ( (rd=(double **)an_malloc_2D(img.x,img.y,sizeof **rd))
			==NULL ) {
		fputs( "error: cannot allocate memory.\n", stderr ) ;
		free( il ) ;
		free( ir ) ;
		fclose( fp ) ;
		return 2 ;
	} else if ( cor3( ir, il, rd, img, fp ) == -1 ) {
		fputs( "error: cannot compute correlation.\n", stderr ) ;
		free( il ) ;
		free( ir ) ;
		fclose( fp ) ;
		return 2 ;
	} else {
		free( il ) ;
		free( ir ) ;
		fclose( fp ) ;
		return 0 ;
	}
}

