#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <an.h>

#define STR_LEN	512
#define UTH	0.5

typedef struct PS {
	int sw ; /* 0: weighted, 1: max, 2: u>u_th */
	int ngrp ;
	AN_IMAGE_COORD img ;
	char fn[STR_LEN] ;
	double *res, *u, *dp, *ur ;
} PS ;

int get_ps( pps, fn )
	PS *pps ;
	char *fn ;
{
	FILE *fp = fopen( fn, "r" ) ;

	if ( fp==NULL )
		return -1 ;
	else {
		int i ;
		char str[STR_LEN] ;

		pps->sw = atoi( fgets( str, STR_LEN, fp ) ) ;

		fgets( pps->fn, STR_LEN, fp ) ;
		for ( i=0; i<STR_LEN && pps->fn[i]!='\0' ; ++i )
			if ( pps->fn[i]=='\n' || pps->fn[i]=='\r' )
				pps->fn[i] = '\0' ;

		pps->dp = (double *)NULL ;
		pps->ngrp = 0 ;
		while( fgets( str, STR_LEN, fp ) != NULL )
			if ( !isdigit(str[0]) )
				continue ;
			else if ( ( pps->dp = (double *)realloc( pps->dp,
				++pps->ngrp*sizeof *pps->dp ) ) == NULL ) {
				free( pps->dp ) ;
				return -1 ;
			} else
				pps->dp[pps->ngrp-1] = atof( str ) ;

		fclose( fp ) ;
		return pps->ngrp ;
	}
}

int put_ps( ps )
	PS ps ;
{
	int i ;

	printf( " -----------\n" ) ;
	printf( " Parameters:\n" ) ;
	printf( " sw: %d\n", ps.sw ) ;
	printf( " file name: %s\n", ps.fn ) ;
	printf( " ngrp=%d\n", ps.ngrp ) ;

	for ( i=0; i<ps.ngrp; ++i )
		printf( " dp=%f\n", ps.dp[i] ) ;

	printf( " -----------\n" ) ;

	return 0 ;
}

int free_mem( ps )
	PS ps ;
{
	free( ps.res ) ;
	free( ps.dp ) ;
	free( ps.u ) ;
	free( ps.ur ) ;

	return 0 ;
}

int all_mem( pps )
	PS *pps ;
{
	int xys = pps->img.x * pps->img.y ;

	pps->res = (double *)malloc( xys*sizeof *pps->res ) ;
	pps->u = (double *)malloc( xys*pps->ngrp*sizeof *pps->u ) ;
	pps->ur = (double *)malloc( xys*sizeof *pps->ur ) ;

	if ( pps->res == NULL || pps->u == NULL || pps->ur == NULL ) {
		free_mem( *pps ) ;
		return -1 ;
	} else
		return 0 ;
}

int get_dpmap_weighted( ps, fn )
	PS ps ;
	char *fn ;
{
	int i, n, t, xys, xysn ;
	double sum ;
	FILE *fpi, *fpo ;

	if ( ( fpi = fopen( ps.fn, "rb" ) ) == NULL )
		return -1 ;
	else if ( ( fpo = fopen( fn, "wb" ) ) == NULL ) {
		fclose( fpi ) ;
		return -1 ;
	} else {
		xys  = ps.img.x * ps.img.y ;
		xysn = xys * ps.ngrp ;
	}

	for ( t=0; t<ps.img.t; ++t ) {
		fread( ps.u, sizeof *ps.u, xysn, fpi ) ;

		for ( i=0; i<xys; ++i ) {
			sum = .0 ;
			ps.res[i] = .0 ;
			for ( n=0; n<ps.ngrp; ++n ) {
				sum += ps.u[n*xys+i] ;
				ps.res[i] += ps.u[n*xys+i]*ps.dp[n] ;
			}
			if ( sum != .0 )
				ps.res[i] /= sum ;
			else
				ps.res[i] = 255. ;
		}

		fwrite( ps.res, sizeof *ps.res, xys, fpo ) ;
	}

	fclose( fpi ) ;
	fclose( fpo ) ;

	return 0 ;
}

int get_dpmap_max( ps, fn )
	PS ps ;
	char *fn ;
{
	int i, n, t, xys, xysn, c ;
	double max_u, max_u2, max_dp, tmp_u, sum_r ;
	FILE *fpi, *fpo, *fpr ;
	char str[STR_LEN] ;

	if ( ( fpi = fopen( ps.fn, "rb" ) ) == NULL )
		return -1 ;
	else if ( ( fpo = fopen( fn, "wb" ) ) == NULL ) {
		fclose( fpi ) ;
		return -1 ;
	} else if ( ( fpr = fopen( strcat( strcpy( str, fn ), ".ur" ), "wb" ) )
			== NULL ) {
		fclose( fpi ) ;
		fclose( fpo ) ;
		return -1 ;
	} else {
		xys  = ps.img.x * ps.img.y ;
		xysn = xys * ps.ngrp ;
	}

	for ( t=0; t<ps.img.t; ++t ) {
		fread( ps.u, sizeof *ps.u, xysn, fpi ) ;

		c = 0 ;
		sum_r = 0. ;
		for ( i=0; i<xys; ++i ) {
			max_u = ps.u[i] ;
			max_u2 = 0. ;
			max_dp = ps.dp[0] ;

			for ( n=0; n<ps.ngrp; ++n ) {
				tmp_u = ps.u[i+n*xys] ;
				if ( tmp_u > max_u ) {
					max_u2 = max_u ;
					max_u = tmp_u ;
					max_dp = ps.dp[n] ;
				} else if ( tmp_u > max_u2 )
					max_u2 = tmp_u ;
			}

			ps.res[i] = max_dp ;
			if ( max_u > 0. ) {
				ps.ur[i] = max_u2/max_u ;
				c++ ;
				sum_r += ps.ur[i] ;
			} else {
				if ( t>0 )
				printf( "max_u=%e at (x=%d,y=%d,t=%d)\n",
					max_u, i%ps.img.x, i/ps.img.x, t ) ;
				ps.ur[i] = .0 ;
			}
		}

		if ( c>0 )
			printf( "t=%d: av_ur=%e\n", t, sum_r/(double)c ) ;
		else
			printf( "t=%d: c=0\n", t ) ;

		fwrite( ps.res, sizeof *ps.res, xys, fpo ) ;
		fwrite( ps.ur, sizeof *ps.ur, xys, fpr ) ;
	}

	fclose( fpi ) ;
	fclose( fpo ) ;
	fclose( fpr ) ;

	return 0 ;
}


int get_dpmap_uth( ps, fn )
	PS ps ;
	char *fn ;
{
	int i, n, t, xys, xysn ;
	double *utmp, sum ;
	FILE *fpi, *fpo ;

	xys = ps.img.x * ps.img.y ;
	xysn = xys * ps.ngrp ;

	if ( ( fpi = fopen( ps.fn, "rb" ) ) == NULL )
		return -1 ;
	else if ( ( fpo = fopen( fn, "wb" ) ) == NULL ) {
		fclose( fpi ) ;
		return -1 ;
	} else if ( ( utmp = (double *)malloc( sizeof *utmp * xysn ) )
				== NULL ) {
		fclose( fpi ) ;
		fclose( fpo ) ;
		return -1 ;
	} else
		for ( i=0; i<xysn; ++i )
			utmp[i] = .0 ;

	for ( t=0; t<ps.img.t; ++t ) {
		fread( ps.u, sizeof *ps.u, xysn, fpi ) ;
		for ( i=0; i<xysn; ++i )
			if ( ps.u[i] > UTH )
				utmp[i] += ps.u[i] ;
/*
			if ( ps.u[i] > UTH )
				utmp[i] += 1. ;
*/
			else
				utmp[i] = 0. ;

		for ( i=0; i<xys; ++i ) {
			sum = .0 ;
			ps.res[i] = .0 ;
			for ( n=0; n<ps.ngrp; ++n ) {
				sum += utmp[n*xys+i] ;
				ps.res[i] += utmp[n*xys+i]*ps.dp[n] ;
			}
			if ( sum != .0 )
				ps.res[i] /= sum ;
			else
				ps.res[i] = 255. ;
		}

		fwrite( ps.res, sizeof *ps.res, xys, fpo ) ;
	}

	free( utmp ) ;
	fclose( fpi ) ;
	fclose( fpo ) ;

	return 0 ;
}

int get_dpmap( ps, fn )
	PS ps ;
	char *fn ;
{
	switch ( ps.sw ) {
		case 0:
			return get_dpmap_weighted( ps, fn ) ;
		case 1:
			return get_dpmap_max( ps, fn ) ;
		case 2:
			return get_dpmap_uth( ps, fn ) ;
		default:
			return -1 ;
	}
}

int main( argc, argv )
	int argc ;
	char **argv ;
{
	PS ps ;

	if ( argc!= 3 ) {
		fprintf( stderr, "Usage: %s data_file result_file.\n", *argv ) ;
		return 0 ;
	} else if ( an_get_img( &ps.img ) == -1 ) {
		fputs( "error: cannot find image environment.\n", stderr ) ;
		return -1 ;
	} else if ( get_ps( &ps, argv[1] ) <= 0 ) {
		fputs( "error: cannot get parameters.\n", stderr ) ;
		return -1 ;
	} else if ( put_ps( ps ) == -1 ) {
		fputs( "error: cannot put parameters.\n", stderr ) ;
		return -1 ;
	} else if ( all_mem( &ps ) == -1 ) {
		fputs( "error: cannot allocate memory.\n", stderr ) ;
		free_mem( ps ) ;
		return -1 ;
	} else if ( get_dpmap( ps, argv[2] ) == -1 ) {
		fputs( "error: cannot get disparity map.\n", stderr ) ;
		free_mem( ps ) ;
		return -1 ;
	} else {
		free_mem( ps ) ;
		return 0 ;
	}
}

